"""
Encoder for categorical variables

(C) 2021-2023 Mengjia Zhu, Alberto Bemporad
"""

import numpy as np
from sklearn.preprocessing import OneHotEncoder

import sys

class cat_encoder:

    def __init__(self, prob):
        """
        Obtain and constructs all necessary attributes for self
        """
        self.nd = prob.nd
        self.X_d = prob.X_d
        self.nci = prob.nci
        self.nci_encoded = prob.nci_encoded

    def cat_encoder(self):
        """
        Generate the one-hot encoder for the categorical variables

        Inputs:
                nd: number of categorical variables to be encoded
                X_d: number of options for each categorical variable
                (for the current setup, assume X_d is known, if its unknown, can be trained from data)
                (which is not the target problem type of this algorithm (experiments are expensive to perform)

        Output:
                encoder: the one-hot encoder for the categorical variables
        """
        nd = self.nd
        X_d = self.X_d
        max_cat = max(X_d)

        xenc_gen = np.array(np.zeros([max_cat, nd]))

        enc = OneHotEncoder(drop=None, sparse=False)

        if nd != len(X_d):
            errstr_nd = "The number of discrete variables is not consistent with the length of X_d (the list indicates number of categorials for each discrete variables), please double check"
            print(errstr_nd)
            sys.exit(1)

        for ind in range(nd):
            x_d_ind = np.arange(X_d[ind])
            xenc_gen[:X_d[ind], ind] = x_d_ind

        encoder = enc.fit(xenc_gen)

        return encoder


    def encode(self, x, encoder):
        """
        Transform the categorical variables in x using the generated one-hot encoder (from 'cat_encoder' function)

        Inputs:
                x: the varaible to be encoded
                encoder: the one-hot encoder generated

        Outputs:
                xencoded: the one-hot encoded version of x

        """

        nci_encoded = self.nci_encoded
        xencoded = x[:, :nci_encoded]
        xtoencode = x[:, nci_encoded:]
        x_d_encoded = encoder.transform(xtoencode)
        xencoded = np.hstack((xencoded, x_d_encoded))

        return xencoded


    def decode(self, x, encoder):
        """
        Reverse one-hot encoding (transform the one-hot encoded variables in x to the original form)

        Inputs:
                x: the varaible to be decoded
                encoder: the one-hot encoder generated

        Outputs:
                xdecoded: the decoded version of x


        """
        nci = self.nci
        xdecoded = x[:, :nci]
        xtodecode = x[:,nci:]
        x_d_decoded = encoder.inverse_transform(xtodecode)
        xdecoded = np.hstack((xdecoded, x_d_decoded))

        return xdecoded


"""
Encoder for integer variables

(C) 2021-2023 Mengjia Zhu, Alberto Bemporad
"""

import numpy as np
from sklearn.preprocessing import OneHotEncoder

import sys

class integ_encoder:

    def __init__(self, prob):
        """
        Obtain and constructs all necessary attributes for self
        """
        self.nint = prob.nint
        self.nc = prob.nc
        self.nci = prob.nci
        self.lb = prob.lb_original
        self.ub = prob.ub_original
        self.int_interval = prob.int_interval
        self.nci_encoded = prob.nci_encoded


    def integ_encoder(self):
        """
        Generate the one-hot encoder for the integer variables

        Inputs:
                nc: number of continuous variables
                nint: number of integer variables to be enoded
                int_interval: number of possible integer values for each integer variable

        Output:
                encoder: the one-hot encoder for the integer variables
        """
        nc = self.nc
        nint = self.nint
        int_interval = self.int_interval
        max_interval = int(round(max(int_interval)))
        lb = self.lb
        ub = self.ub

        xenc_gen = np.array(np.zeros([max_interval, nint]))

        enc = OneHotEncoder(drop=None, sparse=False)

        if nint != int_interval.shape[0]:
            errstr_nint = "The number of integer variables is not consistent with the length of int_interval (the list indicates number of possible integer values for each integer variables), please double check"
            print(errstr_nint)
            sys.exit(1)

        for i in range(nint):
            x_int_ind = np.arange(lb[nc + i], ub[nc + i] + 1).reshape((1,-1))
            xenc_gen[:int(int_interval[i]), i] = x_int_ind

        encoder = enc.fit(xenc_gen)

        return encoder


    def encode(self, x, encoder):
        """
        Transform the integer variables in x using the generated one-hot encoder (from 'integ_encoder' function)

        Inputs:
                x: the varaible to be encoded
                encoder: the one-hot encoder generated

        Outputs:
                xencoded: the one-hot encoded version of x

        """
        nc = self.nc
        nci = self.nci
        xencoded_nc = x[:, :nc]
        xencoded_nd = x[:, nci:]
        xtoencode = x[:, nc:nci]
        x_d_encoded = encoder.transform(xtoencode)
        xencoded = np.hstack((xencoded_nc, x_d_encoded, xencoded_nd))

        return xencoded


    def decode(self, x, encoder):
        """
        Reverse one-hot encoding (transform the one-hot encoded variables in x to the original form)

        Inputs:
                x: the varaible to be decoded
                encoder: the one-hot encoder generated

        Outputs:
                xdecoded: the decoded version of x


        """
        nc = self.nc
        # nci = self.nci
        nci_encoded = self.nci_encoded
        xdecoded_nc = x[:, :nc]
        xdecoded_nd = x[:, nci_encoded:]
        xtodecode = x[:,nc:nci_encoded]
        x_d_decoded = encoder.inverse_transform(xtodecode)
        xdecoded = np.hstack((xdecoded_nc, x_d_decoded, xdecoded_nd))

        return xdecoded


"""
Generate initial samples for the surrogate fitting

(C) 2021-2023 Mengjia Zhu, Alberto Bemporad
"""

import math as mt
from pyDOE import lhs #https://pythonhosted.org/pyDOE/
import pulp as plp
import cdd  #https://pypi.org/project/pycddlib/
import random

from pwasopt.categorical_encoder import *
from pwasopt.integ_encoder import *

class init_sampl:

    def __init__(self, prob):
        """
        Obtain and constructs all necessary attributes for self
        """
        self.isPref = prob.isPref
        self.nsamp = prob.nsamp
        self.maxevals = prob.maxevals
        self.feasible_sampling = prob.feasible_sampling
        self.isLin_eqConstrained = prob.isLin_eqConstrained
        self.Aeq = prob.Aeq
        self.beq = prob.beq
        self.isLin_ineqConstrained = prob.isLin_ineqConstrained
        self.Aineq = prob.Aineq
        self.bineq = prob.bineq

        self.nc = prob.nc
        self.nint = prob.nint
        self.nci = prob.nci
        self.nd = prob.nd
        self.nvars = prob.nvars
        self.nvars_encoded = prob.nvars_encoded
        self.X_d = prob.X_d
        self.sum_X_d = prob.sum_X_d

        self.scale_vars = prob.scale_vars
        self.dd_int = prob.dd_int
        self.d0_int = prob.d0_int

        self.dd_nvars = prob.dd_nvars
        self.d0_nvars = prob.d0_nvars

        self.f = prob.f
        self.lb = prob.lb
        self.ub = prob.ub
        self.lb_nvars = prob.lb_nvars
        self.ub_nvars = prob.ub_nvars
        self.lb_unshrink = prob.lb_unshrink
        self.ub_unshrink = prob.ub_unshrink

        EC_cat = cat_encoder(prob)
        self.encode_cat = EC_cat.encode
        self.decode_cat = EC_cat.decode
        if self.nd > 0:
            self.encoder_cat = EC_cat.cat_encoder()

        self.int_encoded = prob.int_encoded
        self.nint_encoded = prob.nint_encoded
        self.nci_encoded = prob.nci_encoded
        self.int_interval = prob.int_interval
        EC_int = integ_encoder(prob)
        self.encode_int = EC_int.encode
        self.decode_int = EC_int.decode
        if self.int_encoded:
            self.encoder_int = EC_int.integ_encoder()

        self.timelimit = prob.timelimit


    def initial_sampling(self):
        """
        Initial sampling stage

        Outputs:
                Xs: the initial samples obtained, encoded if nd >0
                F: the function evaluation of the samples
                X_samp_decoded: the decoded version of the initial samples
        """

        nsamp = self.nsamp
        maxevals = self.maxevals
        feasible_sampling = self.feasible_sampling
        isLin_eqConstrained = self.isLin_eqConstrained
        Aeq = self.Aeq
        beq = self.beq
        isLin_ineqConstrained = self.isLin_ineqConstrained
        Aineq = self.Aineq
        bineq = self.bineq

        nc = self.nc
        nint = self.nint
        nci = self.nci
        # nci_encoded = self.nci_encoded
        nd = self.nd
        nvars = self.nvars
        nvars_encoded = self.nvars_encoded
        # sum_X_d = self.sum_X_d

        scale_vars = self.scale_vars
        dd_int = self.dd_int
        d0_int = self.d0_int

        f = self.f
        Xs = np.zeros((nsamp, nvars_encoded))
        F = np.zeros((nsamp, 1))

        use_solver = False
        if (not feasible_sampling) or ((not isLin_eqConstrained) and (not isLin_ineqConstrained)):
            # use LHS as initial sampling method
            Xs[0:nsamp, :nvars] = lhs(nvars, nsamp, "m")
            Xs[0:nsamp, :nvars] = Xs[0:nsamp, :nvars] * \
                                 (np.ones((nsamp, 1)) * (self.ub_nvars[:nvars] - self.lb_nvars[:nvars])) + np.ones(
                (nsamp, 1)) * self.lb_nvars[:nvars]

            # for integer variable, round it to the closest integer, then scale it (if scale_vars = True)
            if nint > 0:
                Xs[0:nsamp, nc:nci] = np.round(Xs[0:nsamp, nc:nci])
                if (not self.int_encoded) and (scale_vars):
                    Xs[0:nsamp, nc:nci] = (Xs[0:nsamp, nc:nci] - (np.ones((nsamp, 1)) * d0_int)) * (
                            np.ones((nsamp, 1)) * 1 / dd_int)

            # for binary/categorical variables, round it to the closest integer
            # for binary/categorical variables, generate samples first, then encode (one-hot encoding)
            if nd > 0:
                Xs[0:nsamp, nci:nvars] = np.round(Xs[0:nsamp, nci:nvars])
        else:
            if nint == 0 and nd ==0:
                if isLin_eqConstrained and isLin_ineqConstrained:
                    A_mat = np.vstack((-Aeq, -Aineq, np.identity(nvars_encoded), -np.identity(nvars_encoded)))
                    b_mat = np.vstack((beq, bineq, -self.lb_unshrink.reshape((nvars_encoded, 1)), self.ub_unshrink.reshape((nvars_encoded, 1))))
                elif isLin_eqConstrained:
                    A_mat = np.vstack((-Aeq, np.identity(nvars_encoded), -np.identity(nvars_encoded)))
                    b_mat = np.vstack((beq, -self.lb_unshrink.reshape((nvars_encoded, 1)), self.ub_unshrink.reshape((nvars_encoded, 1))))
                elif isLin_ineqConstrained:
                    A_mat = np.vstack((-Aineq, np.identity(nvars_encoded), -np.identity(nvars_encoded)))
                    b_mat = np.vstack((bineq, -self.lb_unshrink.reshape((nvars_encoded, 1)), self.ub_unshrink.reshape((nvars_encoded, 1))))
                else:
                    A_mat = np.vstack(( np.identity(nvars_encoded), -np.identity(nvars_encoded)))
                    b_mat = np.vstack((-self.lb_unshrink.reshape((nvars_encoded, 1)), self.ub_unshrink.reshape((nvars_encoded, 1))))

                mat_comb = np.hstack((b_mat, A_mat))
                mat = cdd.Matrix(mat_comb)
                mat.rep_type = cdd.RepType.INEQUALITY
                if isLin_eqConstrained:
                    mat.lin_set =[x for x in range(Aeq.shape[0])]
                poly = cdd.Polyhedron(mat)
                gen = poly.get_generators()
                numVert = len(gen)

                list_vert = np.asarray(gen[:])[:, 1:]
                if numVert > nsamp:
                    np.random.shuffle(list_vert)
                    Xs[0:nsamp, :nvars] =list_vert[:nsamp,:]
                else:
                    Xs[0:numVert, :nvars] = list_vert.copy()

                    if not isLin_eqConstrained:
                        coeff_vert = np.random.dirichlet(np.ones(numVert),size = nsamp-numVert)
                        Xs[numVert:nsamp,:nvars] = coeff_vert.dot(list_vert)
                    else:
                        XX_encoded = Xs[:numVert,:].copy()
                        fes_init = 1
                        use_solver = True
                        for ind in range(nsamp-numVert):
                            if self.int_encoded:
                                X_next = self.feasible_sampling_eq_ineq_constrained_intEncoded(XX_encoded, fes_init)
                            else:
                                X_next = self.feasible_sampling_eq_ineq_constrained(XX_encoded, fes_init)
                            Xs[ind+numVert, :] = X_next.copy()

                            XX_encoded = Xs[0:ind + numVert + 1, :]
                            fes_init = 1

            else:
                nn = nsamp
                nk = 0
                n_iter = 0
                n_iter_2 = 0
                while (nk < nsamp):
                    if nn*nvars < 2000 or nk >0:
                        XX = lhs(nvars, nn, "m")
                        XX = XX * (np.ones((nn, 1)) * (self.ub_nvars[:nvars] - self.lb_nvars[:nvars])) + np.ones(
                            (nn, 1)) * self.lb_nvars[:nvars]

                        # for integer variable, round it to the closest integer, then scale it (if scale_vars = True)
                        if nint > 0:
                            XX[:, nc:nci] = np.round(XX[:, nc:nci])
                            if (not self.int_encoded) and (scale_vars):
                                XX[:, nc:nci] = (XX[:, nc:nci] - (np.ones((nn, 1)) * d0_int)) * (np.ones((nn, 1)) * 1 / dd_int)

                        # for binary/categorical variables, rounded it to closest integer
                        # for binary/categorical variables, generate samples first, then encode (one-hot encoding)
                        if nd > 0:
                            XX[:, nci:] = np.round(XX[:, nci:])

                        XX_decoded = XX[:,:nvars].copy()
                        if self.int_encoded and nd > 0:
                            XX_encoded_int = self.encode_int(XX_decoded, self.encoder_int)
                            XX_encoded = self.encode_cat(XX_encoded_int, self.encoder_cat)
                        elif self.int_encoded:
                            XX_encoded = self.encode_int(XX_decoded, self.encoder_int)
                        elif nd > 0:
                            XX_encoded = self.encode_cat(XX_decoded, self.encoder_cat)
                        else:
                            XX_encoded = XX.copy()

                        ii = np.ones((nn, 1)).flatten("C")
                        XX_encoded_fes_check = XX_encoded.copy()
                        if (scale_vars) and (nint > 0) and (not self.int_encoded):
                            XX_encoded_fes_check[:, nc:nci] = XX_encoded_fes_check[:, nc:nci]* (np.ones((nn, 1)) * dd_int) + np.ones((nn, 1)) * d0_int

                        for i in range(nn):
                            if isLin_eqConstrained:
                                ii[i] = all(Aeq.dot(XX_encoded_fes_check[i,].T) <= beq.flatten("c") + 1.0e-8)
                                ii[i] = ii[i] and all(-Aeq.dot(XX_encoded_fes_check[i,].T) <= -beq.flatten("c") + 1.0e-8)
                            if isLin_ineqConstrained:
                                ii[i] = ii[i] and all(Aineq.dot(XX_encoded_fes_check[i,].T) <= bineq.flatten("c") + 1.0e-8)

                        nk = sum(ii)
                    if (n_iter < 4) and (n_iter_2) < 2 and (nk == 0):
                        nn = 20 * nn
                        n_iter_2 += 1
                        if ((isLin_eqConstrained) and (nn > 200)) or ((isLin_ineqConstrained) and (nn > 400)) or (20 * nn >500):
                            n_iter_2 += 1
                        n_iter += 1
                    elif (nk < nsamp) and (nk >0):
                        nn = mt.ceil(min(20, 1.1 * nsamp / nk) * nn)
                        n_iter += 1
                    elif (nk < nsamp) :
                        # use a optimization solver to find initial samples
                        # (when it's hard to find enough # of initial samples that satisfies constraints)
                        use_solver = True

                        if nk >0:
                            ii = np.asarray(np.where(ii))
                            n_fes = ii.size
                            XX = XX[ii[0][0:n_fes],].reshape(n_fes, nvars)
                            fes_init = 1
                        else:
                            n_fes = 0
                            ind_rand = random.randrange(XX.shape[0])
                            XX = XX[ind_rand, :].reshape(1, nvars)
                            fes_init = ii[ind_rand]

                        if self.int_encoded and nd > 0:
                            XX_encoded_int = self.encode_int(XX, self.encoder_int)
                            XX_encoded = self.encode_cat(XX_encoded_int, self.encoder_cat)
                        elif self.int_encoded:
                            XX_encoded = self.encode_int(XX, self.encoder_int)
                        elif nd > 0:
                            XX_encoded = self.encode_cat(XX, self.encoder_cat)
                        else:
                            XX_encoded = XX

                        Xs[0:n_fes,:] = XX_encoded
                        for ind in range(nsamp-n_fes):
                            if self.int_encoded:
                                X_next = self.feasible_sampling_eq_ineq_constrained_intEncoded(XX_encoded, fes_init)
                            else:
                                X_next = self.feasible_sampling_eq_ineq_constrained(XX_encoded, fes_init)
                            Xs[ind+n_fes, :] = X_next.copy()

                            XX_encoded = Xs[0:ind + n_fes + 1, :]
                            fes_init = 1
                        nk = nsamp

        if not use_solver:
            if nint == 0 and nd == 0:
                X_sampl_decoded = Xs[0:nsamp, :]
            elif feasible_sampling and (isLin_ineqConstrained or isLin_eqConstrained):
                ii = np.where(ii)
                Xs[0:nsamp, :nvars] = XX[ii[0][0:nsamp],]
                X_sampl_decoded = Xs[0:nsamp, :nvars].copy()
                if self.int_encoded or nd > 0:
                    X_sampl_encoded = XX_encoded[ii[0][0:nsamp],]
                    Xs[0:nsamp, :] = X_sampl_encoded
            else:
                X_sampl_decoded = Xs[0:nsamp, :nvars].copy()
                if self.int_encoded and nd > 0:
                    X_sampl_encoded_int = self.encode_int(X_sampl_decoded, self.encoder_int)
                    X_sampl_encoded = self.encode_cat(X_sampl_encoded_int, self.encoder_cat)
                elif self.int_encoded:
                    X_sampl_encoded = self.encode_int(X_sampl_decoded, self.encoder_int)
                elif nd > 0:
                    X_sampl_encoded = self.encode_cat(X_sampl_decoded, self.encoder_cat)
                if self.int_encoded or nd > 0:
                    Xs[0:nsamp, :] = X_sampl_encoded
        else:
            X_sampl_decoded = Xs[0:nsamp, :].copy()
            if self.int_encoded and nd > 0:
                X_sampl_decoded_int = self.decode_int(X_sampl_decoded, self.encoder_int)
                X_sampl_decoded = self.decode_cat(X_sampl_decoded_int, self.encoder_cat)
            elif self.int_encoded:
                X_sampl_decoded = self.decode_int(X_sampl_decoded, self.encoder_int)
            elif nd > 0:
                X_sampl_decoded = self.decode_cat(X_sampl_decoded, self.encoder_cat)

        if scale_vars:
            X = list(X_sampl_decoded * (np.ones((nsamp, 1)) * self.dd_nvars) + np.ones((nsamp, 1)) * self.d0_nvars)
        else:
            X = list(X_sampl_decoded)

        # if not self.isPref:
        #     for i in range(nsamp):
        #         F[i] = f(X_sampl_decoded[i, :].T)
        # else:
        #     F = []

        # return Xs, F, X_sampl_decoded
        return Xs, X


    def feasible_sampling_eq_ineq_constrained(self, Xs, fes_init):
        """
        Use MILP to obtain feasible initial samples when integer variables, if exist, are not one-hot encoded
            - Used when the constraints are hard to fulfill by solely randomly sampling
            - Optimization is used to maximize the distance between initial samples

        Inputs:
                Xs: np array, initial samples already obtained
                fes_init: bool,
                    if True, Xs are all feasible
                    if False, Xs are not feasible
                    (necessary for the initial guess of Xs because it was randomly generated)
        Outputs:
                xopt: np array, a feasible initial sample generated

        """

        lb = self.lb
        ub = self.ub
        nc = self.nc
        nint = self.nint
        nci = self.nci
        nd = self.nd
        X_d = self.X_d
        sum_X_d = self.sum_X_d

        scale_vars = self.scale_vars
        dd_int = self.dd_int
        d0_int = self.d0_int

        X_curr_c = Xs[:, :nci]
        X_curr_d = Xs[:, nci:]

        N = np.shape(Xs)[0]

        isLin_eqConstrained = self.isLin_eqConstrained
        Aeq = self.Aeq
        beq = self.beq
        isLin_ineqConstrained = self.isLin_ineqConstrained
        Aineq = self.Aineq
        bineq = self.bineq

        timelimit = self.timelimit

        prob = plp.LpProblem('feas_sampling', plp.LpMinimize)
        xc = plp.LpVariable.dicts("xc", (range(nci), range(1)), cat=plp.LpContinuous)
        xd = plp.LpVariable.dicts("xd", (range(sum_X_d), range(1)), cat=plp.LpBinary)
        xint = plp.LpVariable.dicts("xint", (range(nint), range(1)), cat=plp.LpInteger)
        if (N > 1 or fes_init):
            if nc >0:
                delta_p = plp.LpVariable.dicts("delta_p", (range(N), range(nci)), cat=plp.LpBinary)
                delta_m = plp.LpVariable.dicts("delta_m", (range(N), range(nci)), cat=plp.LpBinary)
            if nint >0:
                delta_p_int = plp.LpVariable.dicts("delta_p_int", (range(N), range(nint)), cat=plp.LpBinary)
                delta_m_int = plp.LpVariable.dicts("delta_m_int", (range(N), range(nint)), cat=plp.LpBinary)
        beta = plp.LpVariable.dicts("beta", range(1), lowBound=0, cat=plp.LpContinuous)
        beta_int = plp.LpVariable.dicts("beta_int",  range(1), lowBound=0, cat=plp.LpContinuous)
        ham_dist_scaled = plp.LpVariable.dicts("ham_dist_scaled", range(1), lowBound=0, cat=plp.LpContinuous)

        # Objective function
        if (N > 1 or fes_init):
            cost = - plp.lpSum(ham_dist_scaled) - plp.lpSum(beta) - plp.lpSum(beta_int) # max the distance between points
        else:
            cost = - plp.lpSum(xc[h][0]-X_curr_c[0][h] for h in range(nci)) - plp.lpSum(xd[h][0]-X_curr_d[0][h] for h in range(sum_X_d))
        prob += cost

        # up and lower bound of xc
        for h in range(nc):
            prob += xc[h][0] <= ub[h]
            prob += xc[h][0] >= lb[h]

        # up and lower bound of xint
        for h in range(nint):
            prob += xint[h][0] <= ub[nc + h]
            prob += xint[h][0] >= lb[nc + h]

        # relationship between xint and xint_scaled
        for h in range(nint):
            if scale_vars:
                prob += xc[nc+h][0]*dd_int[h] + d0_int[h] == xint[h][0]
            else:
                prob += xc[nc + h][0] == xint[h][0]

        # constraints for one-hot encoded xd
        for i in range(nd):
            prob += (plp.lpSum(xd[h][0] for h in range(sum(X_d[:i]), sum(X_d[:i + 1]))) == 1)

        # constraints satisfaction
        if isLin_eqConstrained:
            for i in range(np.size(beq)):
                prob += (plp.lpSum(Aeq[i, h] * xc[h][0] for h in range(nc))
                         + plp.lpSum(Aeq[i, nc + h] * xint[h][0] for h in range(nint))
                         + plp.lpSum(Aeq[i, nci + h] * xd[h][0] for h in range(sum_X_d))
                         == beq[i][0])

        if isLin_ineqConstrained:
            for i in range(np.size(bineq)):
                prob += (plp.lpSum(Aineq[i, h] * xc[h][0] for h in range(nc))
                         + plp.lpSum(Aineq[i, nc + h] * xint[h][0] for h in range(nint))
                         + plp.lpSum(Aineq[i, nci+h] * xd[h][0] for h in range(sum_X_d))
                         <= bineq[i][0])

        # acquisition function related
        if nd > 0 and (N > 1 or fes_init):
            # discrete variables
            sum_ham_dist = 0
            for i in range(N):
                ind_0 = list(np.where(X_curr_d[i, :] == 0)[0])
                ind_1 = list(np.where(X_curr_d[i, :] == 1)[0])
                sum_ham_dist += plp.lpSum(xd[h][0] for h in ind_0) + plp.lpSum(1 - xd[h][0] for h in ind_1)

            prob += ham_dist_scaled == 1 / (nd * N) * sum_ham_dist
        else:
            prob += ham_dist_scaled[0] == 0

        # continuous variables
        # big-M for exploration function
        M_x = 1.0e5 * np.ones((1, nci))
        # if N >1:
        #     max_X_curr = np.zeros((1, nci))
        #     min_X_curr = np.zeros((1, nci))
        #     for h in range(nci):
        #         max_X_curr[0][h] = np.amax(X_curr_c[:, h])
        #         min_X_curr[0][h] = np.amin(X_curr_c[:, h])
        #     for i in range(nci):
        #         if max_X_curr[0,i] > 0:
        #             M_x[0,i] = 2 * max_X_curr[0,i] - min_X_curr[0,i]
        #         else:
        #             M_x[0,i] = max_X_curr[0,i] - 2 * min_X_curr[0,i]

        if nint >0:
            sum_delta_int_N = 0
        if (N > 1 or fes_init):
            for i in range(N):
                if nc >0:
                    for j in range(nc):
                        prob += delta_p[i][j] == 1 - delta_m[i][j]
                        prob += xc[j][0] - X_curr_c[i, j] >= beta[0] - M_x[0, j] * (1 - delta_p[i][j])
                        prob += -xc[j][0] + X_curr_c[i, j] >= beta[0] - M_x[0, j] * (1 - delta_m[i][j])
                    # prob += (plp.lpSum(delta_p[i][j] + delta_m[i][j] for j in range(nci)) >= 1)
                if nint>0:
                    for j in range(nint):
                        prob += delta_p_int[i][j] == 1 - delta_m_int[i][j]
                        prob += (xc[nc + j][0] - X_curr_c[i, nc + j]) * dd_int[j] >= beta_int[0]- M_x[0, nc + j] * (1 - delta_p_int[i][j])
                        prob += (xc[nc + j][0] - X_curr_c[i, nc + j]) * dd_int[j] <= -beta_int[0] + M_x[0, nc + j] * (1 - delta_m_int[i][j])
        else:
            prob += beta[0] == 0
            prob += beta_int[0] == 0

        if nc <1:
            prob += beta[0] == 0
        if nint <1:
            prob += beta_int[0] == 0



        try:
            prob.solve(plp.GUROBI(timeLimit=timelimit, msg=0))
        except:
            prob.solve(plp.GLPK(timeLimit=timelimit, msg=0))
        # prob.solve()
        status = prob.status

        if status == plp.LpStatusOptimal:  # See plp.constants.LpStatus. See more in constants.py in pulp/ folder
            xopt = np.zeros((nci + sum_X_d,))
            for i in range(nci):
                xopt[i,] = xc[i][0].varValue
            for i in range(nci, nci + sum_X_d):
                xopt[i,] = xd[i - nci][0].varValue
        # elif prob.solverModel.status == 9:
        #     print("Acquisition is not solved to optimal, the current best solution is used")
        #     xopt = np.array(prob.solverModel.Xs[:nci+sum_X_d])

            # # # to assess the results:
            # delta_p_opt = np.zeros((N,nci))
            # delta_m_opt = np.zeros((N, nci))
            # for i in range(N):
            #     for j in range(nci):
            #         delta_p_opt[i,j] = delta_p[i][j].varValue
            #         delta_m_opt[i, j] = delta_m[i][j].varValue
            # ham_dist_scaled = ham_dist_scaled[0].varValue
            # beta_opt = beta[0].varValue
        else:
            xopt = None

        return xopt


    def feasible_sampling_eq_ineq_constrained_intEncoded(self, Xs, fes_init):
        """
        Use MILP to obtain feasible initial samples when integer variables, if exist, are one-hot encoded
            - Used when the constraints are hard to fulfill by solely randomly sampling
            - Optimization is used to maximize the distance between initial samples

        Inputs:
                Xs: np array, initial samples already obtained
                fes_init: bool,
                    if True, Xs are all feasible
                    if False, Xs are not feasible
                    (necessary for the initial guess of Xs because it was randomly generated)
        Outputs:
                xopt: np array, a feasible initial sample generated

        """

        lb = self.lb
        ub = self.ub
        nc = self.nc
        nint = self.nint
        # nci = self.nci
        nci_encoded = self.nci_encoded
        nint_encoded = self.nint_encoded
        int_interval = self.int_interval
        nd = self.nd
        X_d = self.X_d
        sum_X_d = self.sum_X_d

        X_curr_c = Xs[:, :nc]
        X_curr_int = Xs[:, nc:nci_encoded]
        X_curr_d = Xs[:, nci_encoded:]

        N = np.shape(Xs)[0]

        isLin_eqConstrained = self.isLin_eqConstrained
        Aeq = self.Aeq
        beq = self.beq
        isLin_ineqConstrained = self.isLin_ineqConstrained
        Aineq = self.Aineq
        bineq = self.bineq

        timelimit = self.timelimit

        prob = plp.LpProblem('feas_sampling', plp.LpMinimize)
        xc = plp.LpVariable.dicts("xc", (range(nc), range(1)), cat=plp.LpContinuous)
        xint = plp.LpVariable.dicts("xint", (range(nint_encoded), range(1)), cat=plp.LpBinary)
        xd = plp.LpVariable.dicts("xd", (range(sum_X_d), range(1)), cat=plp.LpBinary)
        if (N > 1 or fes_init):
            if nc >0:
                delta_p = plp.LpVariable.dicts("delta_p", (range(N), range(nc)), cat=plp.LpBinary)
                delta_m = plp.LpVariable.dicts("delta_m", (range(N), range(nc)), cat=plp.LpBinary)
        beta = plp.LpVariable.dicts("beta", range(1), lowBound=0, cat=plp.LpContinuous)
        int_ham_dist_scaled = plp.LpVariable.dicts("int_ham_dist_scaled", range(1), lowBound=0, cat=plp.LpContinuous)
        cat_ham_dist_scaled = plp.LpVariable.dicts("cat_ham_dist_scaled", range(1), lowBound=0, cat=plp.LpContinuous)

        # Objective function
        if (N > 1 or fes_init):
            cost = - plp.lpSum(cat_ham_dist_scaled) - plp.lpSum(beta) - plp.lpSum(int_ham_dist_scaled) # max the distance between points
        else:
            cost = - plp.lpSum(xc[h][0]-X_curr_c[0][h] for h in range(nc)) \
                   - plp.lpSum(xint[h][0]-X_curr_int[0][h] for h in range(nint_encoded)) \
                   - plp.lpSum(xd[h][0] - X_curr_d[0][h] for h in range(sum_X_d))
        prob += cost

        # up and lower bound of xc
        for h in range(nc):
            prob += xc[h][0] <= ub[h]
            prob += xc[h][0] >= lb[h]

        # constraints for one-hot encoded xd
        for i in range(nd):
            prob += (plp.lpSum(xd[h][0] for h in range(sum(X_d[:i]), sum(X_d[:i + 1]))) == 1)

        # constraints for one-hot encoded xint
        for i in range(nint):
            prob += (plp.lpSum(xint[h][0] for h in range(int(round(sum(int_interval[:i]))), int(round(sum(int_interval[:i + 1]))))) == 1)

        # constraints satisfaction
        if isLin_eqConstrained:
            for i in range(np.size(beq)):
                prob += (plp.lpSum(Aeq[i, h] * xc[h][0] for h in range(nc))
                         + plp.lpSum(Aeq[i, nc + h] * xint[h][0] for h in range(nint_encoded))
                         + plp.lpSum(Aeq[i, nci_encoded + h] * xd[h][0] for h in range(sum_X_d))
                         == beq[i][0])

        if isLin_ineqConstrained:
            for i in range(np.size(bineq)):
                prob += (plp.lpSum(Aineq[i, h] * xc[h][0] for h in range(nc))
                         + plp.lpSum(Aineq[i, nc + h] * xint[h][0] for h in range(nint_encoded))
                         + plp.lpSum(Aineq[i, nci_encoded+h] * xd[h][0] for h in range(sum_X_d))
                         <= bineq[i][0])

        # acquisition function related
        if nd > 0 and (N > 1 or fes_init):
            # discrete variables
            cat_sum_ham_dist = 0
            for i in range(N):
                ind_0 = list(np.where(X_curr_d[i, :] == 0)[0])
                ind_1 = list(np.where(X_curr_d[i, :] == 1)[0])
                cat_sum_ham_dist += plp.lpSum(xd[h][0] for h in ind_0) + plp.lpSum(1 - xd[h][0] for h in ind_1)

            prob += cat_ham_dist_scaled == 1 / (nd * N) * cat_sum_ham_dist
        else:
            prob += cat_ham_dist_scaled[0] == 0

        if nint > 0 and (N > 1 or fes_init):
            # encoded integer variables
            int_sum_ham_dist = 0
            for i in range(N):
                ind_0 = list(np.where(X_curr_int[i, :] == 0)[0])
                ind_1 = list(np.where(X_curr_int[i, :] == 1)[0])
                int_sum_ham_dist += plp.lpSum(xint[h][0] for h in ind_0) + plp.lpSum(1 - xint[h][0] for h in ind_1)

            prob += int_ham_dist_scaled == 1 / (nint * N) * int_sum_ham_dist
        else:
            prob += int_ham_dist_scaled[0] == 0

        # continuous variables
        # big-M for exploration function
        M_x = 1.0e5 * np.ones((1, nc))
        # if N >1:
        #     max_X_curr = np.zeros((1, nc))
        #     min_X_curr = np.zeros((1, nc))
        #     for h in range(nc):
        #         max_X_curr[0][h] = np.amax(X_curr_c[:, h])
        #         min_X_curr[0][h] = np.amin(X_curr_c[:, h])
        #     for i in range(nc):
        #         if max_X_curr[0,i] > 0:
        #             M_x[0,i] = 2 * max_X_curr[0,i] - min_X_curr[0,i]
        #         else:
        #             M_x[0,i] = max_X_curr[0,i] - 2 * min_X_curr[0,i]

        if (N > 1 or fes_init):
            for i in range(N):
                if nc >0:
                    for j in range(nc):
                        prob += delta_p[i][j] == 1 - delta_m[i][j]
                        prob += xc[j][0] - X_curr_c[i, j] >= beta[0] - M_x[0, j] * (1 - delta_p[i][j])
                        prob += -xc[j][0] + X_curr_c[i, j] >= beta[0] - M_x[0, j] * (1 - delta_m[i][j])
                    # prob += (plp.lpSum(delta_p[i][j] + delta_m[i][j] for j in range(nc)) >= 1)
        else:
            prob += beta[0] == 0

        if nc <1:
            prob += beta[0] == 0

        try:
            prob.solve(plp.GUROBI(timeLimit=timelimit, msg=0))
        except:
            prob.solve(plp.GLPK(timeLimit=timelimit, msg=0))
        # prob.solve()
        status = prob.status

        if status == plp.LpStatusOptimal:  # See plp.constants.LpStatus. See more in constants.py in pulp/ folder
            xopt = np.zeros((nci_encoded + sum_X_d,))
            for i in range(nc):
                xopt[i,] = xc[i][0].varValue
            for i in range(nc, nci_encoded):
                xopt[i,] = xint[i - nc][0].varValue
            for i in range(nci_encoded, nci_encoded + sum_X_d):
                xopt[i,] = xd[i - nci_encoded][0].varValue
        # elif prob.solverModel.status == 9:
        #     print("Acquisition is not solved to optimal, the current best solution is used")
        #     xopt = np.array(prob.solverModel.Xs[:nci_encoded+sum_X_d])

            # # # to assess the results:
            # delta_p_opt = np.zeros((N,nc))
            # delta_m_opt = np.zeros((N, nc))
            # for i in range(N):
            #     for j in range(nc):
            #         delta_p_opt[i,j] = delta_p[i][j].varValue
            #         delta_m_opt[i, j] = delta_m[i][j].varValue
            # cat_ham_dist_scaled = cat_ham_dist_scaled[0].varValue
            # beta_opt = beta[0].varValue
        else:
            xopt = None

        return xopt


"""
PWASp -  Preference-based optimization with mixed variables using Piecewise Affine surrogate

[1] M. Zhu and A. Bemporad, “Global and Preference-based Optimization
    with Mixed Variables using Piecewise Affine Surrogates,”
     arXiv preprint arXiv:2302.04686, 2023.

reference code:
                - PARC package by A. Bemporad, 2021, http://cse.lab.imtlucca.it/~bemporad/parc
                - GLIS package by A.Bemporad & M. Zhu, 2023, https://github.com/bemporad/GLIS

(C) 2021-2023 Mengjia Zhu, Alberto Bemporad
"""

from pwasopt.prob_setup import *
from pwasopt.sample import *
from pwasopt.acquisition import *
from pwasopt.categorical_encoder import *
from pwasopt.integ_encoder import *
from pwasopt.fit_surrogate_pwasp import *

import time

class PWASp:
    """
    Main class for PWASp
    Note:
            - The optimization variable are ordered in this way: [continuous vars, integer vars, categorical vars]
            -  When fitting the surrogate, the integer variables (ordinal) are treated as
                - continuous variables if the number of possible integer values are large (see 'int_encoded' in prob_setup.py)
                    - but when query the next point, integer variables are considered as integer variables to enhance constraint satisfication
                - categorical variables, i.e., one-hot encoded if the number of possible integer values are small
                    - the relevant constraints are updated accordingly (see prob_setup.py for more details)
            - Each categorical variable need to be first ordinal encoded as 0 to X_d[i] when specify the problem, then
                - Specify the lower and upper bounds of the categorical variables as follows:
                    - Lower bounds: 0
                    - Upper bounds: X_d[i] -1  (number of possible classes for that caategorical variable -1)
                    - This is to ease the random sample generation in sample.py (generate then encode)
            - If the problem is linearly equality/inequality constrained, when provide the coefficient matrix Aeq/Aineq,
                do not forget to include the columns for EACH options of the categorical variables, if categorical variable exists
            - fit_surrogate_pwasp.py is used to fit the PWA surrogate
            - since numpy array cannot have integer and float variable types together, the user might need to explicitly declare X[nc:nci].dtype(int)
                if strict integer variable is needed.
            - When solve the MILP, it will first attempt to solve via GUROBI, if GUROBI is not available, GLPK will be use
                - User may choose to switch to other solvers by replacing the ones in the acquisition.py
    """

    def __init__(self, pref, lb, ub, delta_E, nc, nint, nd, X_d, nsamp, maxevals, feasible_sampling=True,
                 isLin_eqConstrained=None, Aeq=[], beq=[], isLin_ineqConstrained=None, Aineq=[], bineq=[],
                 K= None, scale_vars = True, shrink_range = True, alpha=1.0e-5, sigma=1, separation=None, maxiter=100, cost_tol=1e-4,
                 min_number=None, fit_on_partition=True, softmax_solver='lbfgs', softmax_maxiter=None, beta=None,
                 initialization=None, verbose=None, categorical=False, timelimit=None, epsDeltaF=None,
                 acq_stage = 'multi-stage', sepvalue=None, synthetic_dm = True):

        """ PWASp - Preference-based optimization with mixed variables using Piecewise Affine surrogate
        (C) Jan, 2023, M.Zhu

        Parameters:
        -----------
        pref: preference indicator
             human or synthetic decision maker who express preferences for two the decision/optimization variables compared
        lb: np array
            lower bounds on the continuous, integer and categorical variables with the following order: [continuous, integers, categorical]
            the lb of categorical variables are 0
        ub: np array
            upper bounds on the continuous, integer and categorical variables with the following order: [continuous, integers, categorical]
            the ub of categorical variables is (the number of categories-1), i.e., X_d[nd_i]-1
        delta_E: float
            the exploration parameter in the acquisition function, which trades-off between the exploitation of the surrogate and the exploration of the exploration function
        nc: int
            number of continuous variables in the optimization variable
        nint: int
            number of integer variables in the optimization variable
        nd: int
            number of categorical variables in the optimization variable
        X_d: list with int elements, dimension: (1 by nd)
            each element in X_d represents the number of options for each categorical variables
            e.g., if nd = 2, nd_1 has 2 possible categories and nd_2 has 4 possible categories, X_d = [2, 4]
        nsamp: int
            number of initial samples
        maxevals: int
            number of maximum function evaluations
        feasible_sampling: bool
            if True, initial samples obtained satisfies the known constraints (if there is any)
            if False, constraints are ignored during initial sampling stage
        isLin_eqConstrained: bool
            if True, the problem has Linear equality constraints
            if False, the problem does not have Linear equality constraints
        (Aeq  = beq)
        Aeq: np array, dimension: (# of linear eq. const by n_encoded), where n_encoded is the length of the optimization variable AFTER being encoded
            the coefficient matrix for the linear equality constraints
        beq: np array, dimension: (n_encode by 1)
            the RHS of the linear eq. constraints
        isLin_ineqConstrained: bool
            if True, the problem has Linear inequality constraints
            if False, the problem does not have Linear inequality constraints
        (Aineq <= bineq)
        Aineq:np array, dimension: (# of linear ineq. const by n_encoded)
            the coefficient matrix for the linear inequality constraints
        bineq: np array, dimension: (n_encode by 1)
            the RHS of the linear ineq. constraints
        K: int
            number of linear affine regressor/classifiers in PWA predictor
        scale_vars: bool
            if True, scale the continuous and integer variables to the range of [-1,1]
            if Flase, scaling is not performed
        shrink_range: bool
            if True, shrink the continuous and integer variables further according to the linear equality constraints
            if Flase, shrink range is not performed

        PARC related parameters, see the definition in PARC package:
            Note: not used in PWASp, placeholder, used in PWAS, see main_pwas.py for more details
            alpha, sigma, separation, maxiter, cost_tol, min_number, fit_on_partition, softmax_solver, softmax_maxiter, beta,
                initialization, verbose, categorical

        timelimit: float
            maximum time allowed for the solver in PULP to solve the MILP/LP/MIP
        epsDeltaF: float
            the tolerance for the difference bewtten function evaluations
        acq_stage: str
            if nd>0, weather to solve the MILP in one-stage or two stage (solve MIP first, then LP)
            either 'one-stage' or 'two-stage'

        sepvalue: float
            the value used in constructing the surrogate function fhat:
            fhat(x1)<=fhat(x2)-sepvalue if pref(x1,x2) = -1
            |fhat(x1)-fhat(x2)|<=sepvalue if pref(x1,x2) = 0.
        synthetic_dm: bool
            if True, a synthetic decision maker is used
                - a fun act as the decision maker whose evaluation and feasibility are used to express preferences
                - Note: the actual evaluations of the fun. are unknown to the PWASp solver
            if False, other decision maker is used

        """

        # obtain the problem setup
        isPref = 1
        self.prob = problem_defn(isPref, pref, lb, ub, delta_E, nc, nint, nd, X_d, nsamp, maxevals, feasible_sampling,
                                 isLin_eqConstrained, Aeq, beq, isLin_ineqConstrained, Aineq, bineq,
                                 K, scale_vars,shrink_range,  alpha, sigma, separation, maxiter, cost_tol, min_number,
                                 fit_on_partition, softmax_solver, softmax_maxiter, beta, initialization,
                                 verbose, categorical, timelimit, epsDeltaF, acq_stage, sepvalue, synthetic_dm)

        # obtain the encoder if categorical variables are involved
        if nd > 0:
            EC_cat = cat_encoder(self.prob)
            self.encoder_cat = EC_cat.cat_encoder()

        # obtain the encoder if integer variables are encoded
        if self.prob.int_encoded:
            EC_int = integ_encoder(self.prob)
            self.encoder_int = EC_int.integ_encoder()

        self.AL = active_learn(self.prob)

        self.isInitialized = False
        self.X = list()  # decision variable in the original format (X in the paper)
        self.Xs = list()  # the scaled and/or encoded decision vector (\bar X in the paper)
        self.iter = 0
        self.xnext = None
        self.xsnext = None
        self.fbest = np.inf
        self.ibest = None
        self.xbest = None
        self.xsbest = None
        self.ibest_seq = list()  # keep track of the index of the current best decision vector
        self.isfeas_seq = list()  # keep track of the feasibility of the tested decision vectors
        self.time_fun_eval = list()  # keep track of the CPU time used to evaluation one function/simulation/experiment
        self.time_opt_acquisition = list() # keep track of the CPU time used to optimize the acq. fun. at each iteration
        self.time_fit_surrogate = list() # keep track of the CPU time used to fit the surrogate fun. at each iteration
        self.I = list()
        self.Ieq = list()

    def isKnownFeasible(self, xs):
        """
        Check the feasibility of sample xs w.r.t known constraints
        """
        Aeq = self.prob.Aeq
        beq = self.prob.beq
        Aineq = self.prob.Aineq
        bineq = self.prob.bineq

        isfeas = True
        if self.isLin_eqConstrained:
            isfeas = isfeas and all(Aeq @ xs <= beq + 1.0e-8)
            isfeas = isfeas and all(-Aeq @ xs <= -beq + 1.0e-8)
        if self.isLin_ineqConstrained:
            isfeas = isfeas and all(Aineq @ xs <= bineq+ 1.0e-8)
        return isfeas


    def initialize(self):
        """
        Initialize the problem
            - obtain the initial samples to query

        return:
            self.xnext: 1D array
                the initial sample to query
        """
        s = init_sampl(self.prob)
        Xs, X = s.initial_sampling()
        self.X = X
        self.Xs = Xs
        # we arbitrarily consider the current best = first sample Xs[0] and
        # will ask for comparing Xs[0] wrt Xs[1] next:
        self.xbest = self.X[0]
        self.xsbest = self.Xs[0]
        self.xnext = self.X[1]
        self.xsnext = self.Xs[1]
        self.isInitialized = True
        if self.prob.nint > 0:
            self.xbest[self.prob.nc:self.prob.nci] = np.round(
                self.xbest[self.prob.nc:self.prob.nci])
            self.xnext[self.prob.nc:self.prob.nci] = np.round(
                self.xnext[self.prob.nc:self.prob.nci])

        self.ibest = 0
        self.ibest_seq.append(self.ibest)
        self.iter = 1

        return self.xbest, self.xnext, self.xsbest, self.xsnext

    def update(self, pref_val):
        """
        - Update the relevant variables w.r.t the newly queried sample
        - And then solve the optimization problem on the updated acquisition function to obtain the next point to query

        - Note:
            - initial samples are always feasible wrt known constraints if self.prob.feasible_sampling = True
            - actively generated samples are always feasible wrt known constraints (constraints are enforced in the MILP)

        Input:
            pref_val: int
                pairwise comparison result w.r.t x (the last queried point) and current best point
        Return:
            self.xnext: 1D-array
                The next point to query
        """
        nc = self.prob.nc
        nint = self.prob.nint
        nint_encoded = self.prob.nint_encoded
        nci = self.prob.nci
        nci_encoded = self.prob.nci_encoded
        nd = self.prob.nd
        if nd > 0:
            EC_cat = cat_encoder(self.prob)
        int_encoded = self.prob.int_encoded
        if int_encoded:
            EC_int = integ_encoder(self.prob)
        nvars_encoded = self.prob.nvars_encoded
        sum_X_d = self.prob.sum_X_d

        acq_stage = self.prob.acq_stage

        x = self.xnext # this was either computed at the previous call after n_initial_random iterations or iterated from the initial samples
        xs = self.xsnext
        N = self.iter  # current sample being examined

        if self.iter < self.prob.nsamp:
            isfeas = True
            if not self.prob.feasible_sampling:
                isfeas = self.isKnownFeasible(xs)
            self.time_opt_acquisition.append(0.)
            self.time_fit_surrogate.append(0.)
        else:
            isfeas = True  # actively generated samples are always feasible wrt known constraints

        self.isfeas_seq.append(isfeas)

        if pref_val == -1:
            self.I.append([N, self.ibest])
            self.xbest = x.copy()
            self.xsbest = xs.copy()
            self.ibest = N
        elif pref_val == 1:
            self.I.append([self.ibest, N])
        else:
            self.Ieq.append([N, self.ibest])
        self.ibest_seq.append(self.ibest)

        if self.prob.verbose > 0:
            if self.ibest == N:
                txt = '(***improved x!)'
            else:
                txt = '(no improvement)'
            self.results_display(N, x, self.ibest + 1, txt)

        if N >= self.prob.nsamp-1:
            # Active sampling

            # Assignment of points to PWL
            X_curr = self.Xs.copy()
            t0 = time.time()
            FS = fit_surrogate(self.prob)
            delta = FS.get_init_delta(X_curr, N+1)
            Kf, delta, omega, gamma = FS.get_pwl_param(X_curr, delta, N+1)
            a, b, y_pred = FS.get_parameters(X_curr, np.array(self.I), np.array(self.Ieq), N+1, Kf, delta)
            self.time_fit_surrogate.append(time.time() - t0)

            dF = max(max(y_pred) - min(y_pred), self.prob.epsDeltaF)

            t0 = time.time()
            skip_z = False
            if acq_stage == 'multi-stage':
                if nd > 0:
                    z1 = self.AL.discrete_explore(X_curr[:, nci_encoded:], self.xsbest[:nci_encoded].reshape(nci_encoded, 1), a, b,
                                             N, omega, gamma, Kf, dF)
                    if np.isnan(z1).any():
                        print(
                            'The optimal solution is not reached within the timeLimit, solution is sampled using \'acq_surrogate\' function in acquisition.py.')
                        if int_encoded:
                            z = self.AL.acq_surrogate_intEncoded(a, b, omega, gamma, Kf, dF)
                        else:
                            z = self.AL.acq_surrogate(a, b, omega, gamma, Kf, dF)
                        skip_z = True
                else:
                    z1 = np.array([])
                if not skip_z:
                    z = z1

                if nint > 0:
                    if not skip_z:
                        if int_encoded:
                            z2 = self.AL.integ_explore_intEncoded(X_curr[:, nc:nci_encoded], self.xsbest[:nc].reshape(nc, 1),
                                                             z1.reshape(sum_X_d, 1), a,
                                                             b, N, omega, gamma, Kf, dF)
                        else:
                            z2 = self.AL.integ_explore(X_curr[:, nc:nci_encoded], self.xsbest[:nc].reshape(nc, 1),
                                                  z1.reshape(sum_X_d, 1), a, b, N, omega, gamma, Kf, dF)
                            if np.isnan(z2).any():
                                print(
                                    'The optimal solution is not reached within the timeLimit, solution is sampled using \'acq_surrogate\' function in acquisition.py.')
                                if int_encoded:
                                    z = self.AL.acq_surrogate_intEncoded(a, b, omega, gamma, Kf, dF)
                                else:
                                    z = self.AL.acq_surrogate(a, b, omega, gamma, Kf, dF)
                                skip_z = True
                else:
                    z2 = np.array([])
                if not skip_z:
                    z = np.hstack((z2, z))

                if nc > 0:
                    if not skip_z:
                        z3 = self.AL.cont_explore(X_curr[:, :nc], z2.reshape(nint_encoded, 1), z1.reshape(sum_X_d, 1), a, b,
                                             N, omega, gamma, Kf, dF)
                        if np.isnan(z3).any():
                            print(
                                'The optimal solution is not reached within the timeLimit, solution is sampled using \'acq_surrogate\' function in acquisition.py.')
                            if int_encoded:
                                z = self.AL.acq_surrogate_intEncoded(a, b, omega, gamma, Kf, dF)
                            else:
                                z = self.AL.acq_surrogate(a, b, omega, gamma, Kf, dF)
                            skip_z = True
                else:
                    z3 = np.array([])
                if not skip_z:
                    z = np.hstack((z3, z))

            elif acq_stage == 'one-stage':
                # Solve the acquisition step in one stage
                if int_encoded:
                    z = self.AL.acq_explore_intEncoded(X_curr, a, b, N, omega, gamma, Kf, dF)
                else:
                    z = self.AL.acq_explore(X_curr, a, b, N, omega, gamma, Kf, dF)
                if np.isnan(z).any():
                    print(
                        'The optimal solution is not reached within the timeLimit, solution is sampled using \'acq_surrogate\' function in acquisition.py.')
                    z = self.AL.acq_surrogate(a, b, omega, gamma, Kf, dF)

            else:
                errstr_acq_stage = "acq_stage can only be 'one-stage' or 'multi-stage', please check the string assigned"
                print(errstr_acq_stage)
                sys.exit(1)
            self.time_opt_acquisition.append(time.time() - t0)

            z_decoded = z.copy()
            if int_encoded and nd > 0:
                z_decoded_int = EC_int.decode(z.reshape(1, nvars_encoded), self.encoder_int)
                z_decoded = EC_cat.decode(z_decoded_int, self.encoder_cat)
            elif int_encoded:
                z_decoded = EC_int.decode(z.reshape(1, nvars_encoded), self.encoder_int)
            elif nd > 0:
                z_decoded = EC_cat.decode(z.reshape(1, nvars_encoded), self.encoder_cat)

            self.xsnext = z.T.reshape(nvars_encoded)
            self.Xs = np.vstack((self.Xs, self.xsnext))
            self.xnext = z_decoded.T.reshape(self.prob.nvars) * self.prob.dd_nvars + self.prob.d0_nvars
            if nint > 0:
                self.xnext[nc:nci] = np.round(self.xnext[nc:nci])

            self.X.append(self.xnext)

        else:
            self.xnext = self.X[self.iter + 1]
            self.xsnext = self.Xs[self.iter + 1]

        self.iter += 1
        return self.xnext


    def solve(self):
        """
        If the pref_fun (decision-maker process) have already be integrated with the PWASp solver,
            - use solve() to solve the problem directly

        Return:
            self.xbest: 1D-array
                the best x sampled
        """
        t_all = time.time()

        xbest, x, xsbest, xs = self.initialize()  # x, self.xbest are unscaled. Initially, xbest is always the first random sample

        pref = self.prob.pref
        if self.prob.synthetic_dm:
            self.prob.pref_fun.clear()

        for k in range(self.prob.maxevals-1):
            t0 = time.time()
            pref_val = pref(x, xbest, xs, xsbest)
            self.time_fun_eval.append(time.time() - t0)
            x = self.update(pref_val)
            xbest = self.xbest

        self.X = self.X[:-1]  # it is because in prob.update, it will calculate the next point to query (the last x2 is calculated at max_evals +1)
        self.time_total = time.time() - t_all
        return self.xbest


    def results_display(self, N, z, ibest,txt):
        # Display intermediate results

        z = z.reshape(self.prob.nci + self.prob.nd)
        string = ("N = %4d %s: x = [" % (N,txt))

        for j in range(self.prob.nc):
            aux = z[j]
            if self.prob.scale_vars:
                aux = aux * self.prob.dd[j] + self.prob.d0[j]
            string = string + " x" + str(j + 1) + " = " + ('%7.4f' % aux) + "   "
        for j in range(self.prob.nint):
            aux = z[self.prob.nc+j]
            if self.prob.scale_vars and (not self.prob.int_encoded):
                aux = aux * self.prob.dd_int[j] + self.prob.d0_int[j]
            string = string + " x" + str(j + self.prob.nc + 1) + " = " + ('%5d' % aux) + "   "
        for j in range(self.prob.nd):
            aux = z[self.prob.nci+j]
            string = string + " x" + str(j + self.prob.nci + 1) + " = " + ('%5d' % aux)

        string = string + "]," + "   " + "N_best" + "=" + ('%5d' % ibest)
        print(string)
        return















"""
Function used to express preferences (for numerical benchmarks)

Note: the exact numerical results are assumed to be unknown and only used to express preferences

Reference code: 'glis_function1' version 3.0 in glis package by A. Bemporad, 2021
                http://cse.lab.imtlucca.it/~bemporad/glis/index.html

"""

from numpy import vstack, append, array
from numpy import sum as npsum
import numpy as np

class PWASp_fun1:
    """
    preference query function for numerical benchmark

    Reference code: 'glis_function1' version 3.0 in glis package by A. Bemporad, 2021
    """

    def __init__(self, f, comparetol, Aeq,beq,Aineq, bineq):
        self.itest = 0
        self.Xtest = []
        self.Ftest = []
        self.Festest = []
        self.f = f
        self.comparetol = comparetol
        self.Aeq = Aeq
        self.beq = beq
        self.Aineq = Aineq
        self.bineq = bineq

    def clear(self):
        self.itest = 0
        self.Xtest = []
        self.Ftest = []
        self.Festest = []
        return

    def eval(self, x, y, x_encoded,y_encoded):

        xfound = False
        yfound = False

        itest = self.itest
        if itest > 0:
            Xtest = self.Xtest
            Ftest = self.Ftest
            Festest = self.Festest
        else:
            fx = self.f(x)
            fes_known_x = self.feas_check(x_encoded)
            itest = 1
            Xtest = array([x])
            Ftest = array([fx])
            Festest = array([fes_known_x])
            xfound = True

        for i in range(itest):
            if not (xfound) and npsum(abs(Xtest[i, :] - x)) <= 1e-10:
                xfound = True
                fx = Ftest[i]
                fes_known_x = Festest[i]

            if not (yfound) and npsum(abs(Xtest[i, :] - y)) <= 1e-10:
                yfound = True
                fy = Ftest[i]
                fes_known_y = Festest[i]

        if not (xfound):
            fx = self.f(x)
            fes_known_x = self.feas_check(x_encoded)
            Xtest = vstack((Xtest, x))
            Ftest = append(Ftest, fx)
            Festest = append(Festest, fes_known_x)
            itest = itest + 1

        if not (yfound):
            fy = self.f(y)
            fes_known_y = self.feas_check(y_encoded)
            Xtest = vstack((Xtest, y))
            Ftest = append(Ftest, fy)
            Festest = append(Festest, fes_known_y)
            itest = itest + 1

        # Make comparison
        if fx < fy - self.comparetol:
            if fes_known_x == 1 or (fes_known_y == 0 and fes_known_x == 0):
                out = -1
            else:
                out = 1
        elif fx > fy + self.comparetol:
            if fes_known_y == 1 or (fes_known_y == 0 and fes_known_x == 0):
                out = 1
            else:
                out = -1
        else:
            out = 0

        self.Xtest = Xtest
        self.Ftest = Ftest
        self.Festest = Festest
        self.itest = itest

        return out

    def value(self, x):
        # Compute function value, from available ones if available
        #
        # (C) 2019 A. Bemporad, September 22, 2019

        j = 0
        while j < self.itest:
            if npsum(abs(self.Xtest[j, :] - x)) <= 1e-10:
                val = self.Ftest[j]
                return val
            j = j + 1

        # Value does not exist, compute it
        val = self.f(x)
        self.Xtest = vstack((self.Xtest, x))
        self.Ftest = append(self.Ftest, val)
        self.itest = self.itest + 1
        return val

    def feas_check(self, x):
        # check if the decision variable is feasible or not ( for known constraints, which is needed to help express preferences)
        isfeas = True
        if self.Aineq.any():
            isfeas = isfeas and all(self.Aineq.dot(x) <= self.bineq.flatten("c")+ 1.0e-8)
        if self.Aeq.any():
            isfeas = isfeas and all(self.Aeq.dot(x) <= self.beq.flatten("c") + 1.0e-8)
            isfeas = isfeas and all(-self.Aeq.dot(x) <= -self.beq.flatten("c") + 1.0e-8)

        return isfeas


"""
Set up the problem

(C) 2021-2023 Mengjia Zhu, Alberto Bemporad
"""

import numpy as np
from scipy.optimize import linprog as linprog
import sys

from pwasopt.pref_fun1 import PWASp_fun1
from pwasopt.pref_fun import PWASp_fun


class problem_defn:
    """
    Initial set up of the problem (problem description/definition)

    """

    def __init__(self, isPref, fun, lb, ub, delta_E, nc, nint, nd, X_d, nsamp, maxevals, feasible_sampling,
                 isLin_eqConstrained, Aeq, beq, isLin_ineqConstrained, Aineq, bineq, K, scale_vars, shrink_range,
                 alpha, sigma, separation, maxiter, cost_tol, min_number,
                 fit_on_partition, softmax_solver, softmax_maxiter, beta, initialization,
                 verbose, categorical,timelimit, epsDeltaF, acq_stage, sepvalue, synthetic_dm):

        self.synthetic_dm = synthetic_dm
        self.isPref = isPref

        self.delta_E = delta_E
        self.nc = nc
        self.nint = nint
        self.nci = nc + nint
        self.nd = nd
        self.X_d = X_d
        self.sum_X_d = int(round(sum(X_d)))
        self.nvars = self.nci + nd
        self.nvars_encoded = self.nci + self.sum_X_d

        if np.isinf(lb).any() or np.isinf(ub).any():
            errstr_inf = "Please specify non-infinity upper and lower bounds"
            print(errstr_inf)
            sys.exit(1)

        if lb.shape[0] != self.nvars or ub.shape[0] != self.nvars:
            errstr_bd_shape = "Please specify the lower and upper bounds for all types of variables. Check the Notes in 'main_pwas.py' or 'main_pwasp for how to specify the bounds for the categorical variables"
            print(errstr_bd_shape)
            sys.exit(1)

        lb_encoded = np.zeros(self.nvars_encoded)
        ub_encoded = np.ones(self.nvars_encoded)
        lb_encoded[:self.nci] = lb[:self.nci]
        ub_encoded[:self.nci] = ub[:self.nci]
        self.lb = lb_encoded
        self.ub = ub_encoded
        self.lb_nvars = lb.copy()
        self.ub_nvars = ub.copy()
        self.lb_original = lb.copy()
        self.ub_original = ub.copy()

        self.nsamp = nsamp
        self.maxevals = maxevals

        if self.maxevals < self.nsamp:
            errstr = "Max number of function evaluations is too low. You specified"
            errstr = errstr + " maxevals = " + str(maxevals) + " and nsamp = " + str(nsamp)
            print(errstr)
            sys.exit(1)

        self.feasible_sampling = feasible_sampling
        self.isLin_eqConstrained = isLin_eqConstrained
        self.Aeq = Aeq
        self.beq = beq
        self.isLin_ineqConstrained = isLin_ineqConstrained
        self.Aineq = Aineq
        self.bineq = bineq

        if K is None:
            errstr_k = "Please specify the number of initial partitions"
            print(errstr_k)
            sys.exit(1)
        else:
            self.K = K

        if nc ==0:
            scale_vars = False

        self.scale_vars = scale_vars

        dd = np.ones((self.nvars_encoded,)) # To be used for constraints scaling
        d0 = np.zeros((self.nvars_encoded,))
        dd[:nc] = (ub[:nc] - lb[:nc]) / 2
        d0[:nc] = (ub[:nc] + lb[:nc]) / 2
        # Note: in dd and d0, for integer variables, no scaling is applied (when apply constraints scaling, integer variables are not scaled)
        # instead, in the following, dd_int and d0_int are used, when int variables need to be scaled
        self.dd = dd
        self.d0 = d0

        self.dd_int = (ub[nc:self.nci] - lb[nc:self.nci]) / 2
        self.d0_int = (ub[nc:self.nci] + lb[nc:self.nci]) / 2

        self.alpha = alpha
        self.sigma = sigma

        if separation is None:
            self.separation = 'Softmax'
        else:
            self.separation = separation

        self.maxiter = maxiter
        self.cost_tol = cost_tol

        if min_number is None:
            self.min_number = 1
        else:
            self.min_number = min_number

        self.fit_on_partition = fit_on_partition
        self.softmax_solver = softmax_solver

        if softmax_maxiter is None:
            self.softmax_maxiter = 10000
        else:
            self.softmax_maxiter = softmax_maxiter

        if beta is None:
            self.beta = 0.02
        else:
            self.beta = beta

        if initialization is None:
            self.initialization = ("kmeans", 10)
        else:
            self.initialization = initialization

        if verbose is None:
            self.verbose = 1
        else:
            self.verbose = verbose

        if timelimit is None:
            self.timelimit = self.nvars_encoded * 3  # seconds
            # self.timelimit = 5
        else:
            self.timelimit = timelimit

        if epsDeltaF is None:
            self.epsDeltaF = 1e-4
        else:
            self.epsDeltaF = epsDeltaF

        self.categorical = categorical
        self.acq_stage = acq_stage

        if sepvalue is None:
            self.sepvalue = float(1/maxevals)
        else:
            self.sepvalue = float(sepvalue)

        if self.scale_vars:
            lb_encoded[:nc] = -np.ones(nc)
            ub_encoded[:nc] = np.ones(nc)
            self.lb = lb_encoded
            self.ub = ub_encoded
            self.lb_nvars[:nc] = -np.ones(nc)
            self.ub_nvars[:nc] = np.ones(nc)

            if isLin_eqConstrained:
                if Aeq.shape[1] != self.nvars_encoded:
                    errstr_eq = "The size of the linear equality constraint matrix is not consistent with the encoded optimization variables (please include an INDIVIDUAL constraint for each option for categorical/discrete/binary varialbes)"
                    print(errstr_eq)
                    sys.exit(1)
                self.Aeq = Aeq.dot(np.diag(dd.flatten('C')))
                self.beq = beq-Aeq.dot(d0.reshape(self.nvars_encoded,1))

            if isLin_ineqConstrained:
                if Aineq.shape[1] != self.nvars_encoded:
                    errstr_eq = "The size of the linear inequality constraint matrix is not consistent with the encoded optimization variables (please include an INDIVIDUAL constraint for each option for categorical/discrete/binary varialbes)"
                    print(errstr_eq)
                    sys.exit(1)
                self.Aineq = Aineq.dot(np.diag(dd.flatten('C')))
                self.bineq = bineq - Aineq.dot(d0.reshape(self.nvars_encoded,1))

        self.lb_unshrink = self.lb.copy()
        self.ub_unshrink = self.ub.copy()
        if (shrink_range) and (isLin_ineqConstrained):
            lb_shrink = self.lb.copy()
            ub_shrink = self.ub.copy()
            bounds_linprog = np.zeros((self.nvars_encoded,2))
            bounds_linprog[:,0] = self.lb.copy()
            bounds_linprog[:, 1] = self.ub.copy()
            flin = np.zeros((self.nvars_encoded,1))
            for i in range(self.nvars_encoded):
                flin[i] = 1
                res = linprog(flin,self.Aineq,self.bineq,bounds = bounds_linprog)
                aux = max(self.lb[i],res.fun)
                lb_shrink[i] = aux

                flin[i] = -1
                res = linprog(flin,self.Aineq,self.bineq,bounds =bounds_linprog)
                aux = min(self.ub[i],-res.fun)
                ub_shrink[i] = aux
                flin[i] = 0

            self.lb = lb_shrink
            self.ub = ub_shrink

            self.lb_nvars[:self.nci] = lb_shrink[:self.nci]
            self.ub_nvars[:self.nci] = ub_shrink[:self.nci]

        if nint >0:
            int_interval = np.round(self.ub_original[self.nc:self.nci] - self.lb_original[self.nc:self.nci] + 1)
            self.int_prod = np.prod(int_interval)
            int_sum = int(round(np.sum(int_interval)))
            self.int_interval = int_interval
        else:
            self.int_prod = 0
            self.int_interval = []

        if nint >0 and self.int_prod < maxevals:
            self.int_encoded = True
            # if number of possible combinations of integer variables exceed the number of max evaluations, one-hot encode integer variables
            if isLin_eqConstrained:
                Aeq_int_encode = np.zeros((Aeq.shape[0],nc+int_sum+self.sum_X_d))
                Aeq_int_encode[:,:nc] = self.Aeq[:,:nc]
                Aeq_int_encode[:,nc+int_sum:] = self.Aeq[:,self.nci:]
                for i in range(nint):
                    Aeq_int_encode[:, nc + int(round(np.sum(int_interval[:i]))):nc + int(round(np.sum(int_interval[:i + 1])))] = \
                        self.Aeq[:,nc+i].dot(np.arange(self.lb[nc+i],self.ub[nc+i]+1))

                self.Aeq = Aeq_int_encode

            if isLin_ineqConstrained:
                Aineq_int_encode = np.zeros((Aineq.shape[0], nc + int_sum + self.sum_X_d))
                Aineq_int_encode[:, :nc] = self.Aineq[:, :nc]
                Aineq_int_encode[:, nc + int_sum:] = self.Aineq[:, self.nci:]
                for i in range(nint):
                    Aineq_int_encode[:, nc + int(round(np.sum(int_interval[:i]))):nc + int(round(np.sum(int_interval[:i + 1])))] = \
                        self.Aineq[:,nc+i].reshape((-1,1)).dot(np.arange(self.lb_original[nc + i], self.ub_original[nc + i] + 1).reshape((1,-1)))

                self.Aineq = Aineq_int_encode

            self.nint_encoded = int_sum
            self.nci_encoded = self.nc + self.nint_encoded
            self.nvars_encoded = self.nci_encoded + self.sum_X_d

            lb_int_encoded = np.zeros(int(round(self.nvars_encoded)))
            ub_int_encoded = np.ones(int(round(self.nvars_encoded)))
            lb_int_encoded[:self.nc] = self.lb[:self.nc]
            ub_int_encoded[:self.nc] = self.ub[:self.nc]
            self.lb = lb_int_encoded
            self.ub = ub_int_encoded
        else:
            self.int_encoded = False
            self.nint_encoded = nint
            self.nci_encoded = self.nci

        dd_nvars = np.ones((self.nvars,)) # Here, integer variables are also scaled, to be used for obj. fun. scaling
        d0_nvars = np.zeros((self.nvars,))
        if self.int_encoded:
            dd_nvars[:self.nc] = (ub[:self.nc] - lb[:self.nc]) / 2
            d0_nvars[:self.nc] = (ub[:self.nc] + lb[:self.nc]) / 2
        else:
            dd_nvars[:self.nci] = (ub[:self.nci] - lb[:self.nci]) / 2
            d0_nvars[:self.nci] = (ub[:self.nci] + lb[:self.nci]) / 2
        self.dd_nvars = dd_nvars
        self.d0_nvars = d0_nvars

        dd_nvars_encoded_ = np.ones((self.nvars_encoded,)) # Here, integer variables are also scaled
        d0_nvars_encoded_ = np.zeros((self.nvars_encoded,))
        if self.int_encoded:
            dd_nvars_encoded_[:self.nc] = dd_nvars[:self.nc]
            d0_nvars_encoded_[:self.nc] = d0_nvars[:self.nc]
        else:
            dd_nvars_encoded_[:self.nci] = dd_nvars[:self.nci]
            d0_nvars_encoded_[:self.nci] = d0_nvars[:self.nci]


        if isPref:
            if synthetic_dm:
                comparetol = 1e-4
                if isLin_ineqConstrained or isLin_eqConstrained:
                    pref_fun = PWASp_fun1(fun, comparetol, self.Aeq, self.beq, self.Aineq, self.bineq)  # preference function object
                else:
                    pref_fun = PWASp_fun(fun, comparetol)
                pref = lambda x, y, x_encoded, y_encoded: pref_fun.eval(x, y, x_encoded, y_encoded)
                self.pref = pref
                self.pref_fun = pref_fun
            else:
                self.pref = fun
            self.f = lambda x: 0
        else:
            self.f = fun
            self.pref = lambda x: 0

        # if self.scale_vars:
        #     if isPref:
        #         self.pref = lambda x, y, x_encoded, y_encoded: pref(x * self.dd_nvars + self.d0_nvars,
        #                                                            y * self.dd_nvars + self.d0_nvars,
        #                                                            x_encoded * dd_nvars_encoded_ + d0_nvars_encoded_,
        #                                                            y_encoded * dd_nvars_encoded_ + d0_nvars_encoded_)
        #     else:
        #         self.f = lambda x: fun(x * self.dd_nvars + self.d0_nvars)






